-- author: Raphael Greshake
-- date: 19.10.2019

-- do not use this script in other mods without my permission
-- note: a bale position is only active if a bale is attached to it

BaleCollector = {}

function BaleCollector.prerequisitesPresent(specializations)
    return true;
end;

function BaleCollector.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "baleTriggerCallback", BaleCollector.baleTriggerCallback);
	SpecializationUtil.registerFunction(vehicleType, "getIsBaleAllowed", BaleCollector.getIsBaleAllowed);
	SpecializationUtil.registerFunction(vehicleType, "getNearestBalePosition", BaleCollector.getNearestBalePosition);
	SpecializationUtil.registerFunction(vehicleType, "addDynamicMountedObject", BaleCollector.addDynamicMountedObject)
    SpecializationUtil.registerFunction(vehicleType, "removeDynamicMountedObject", BaleCollector.removeDynamicMountedObject)
    SpecializationUtil.registerFunction(vehicleType, "collectBales", BaleCollector.collectBales)
    SpecializationUtil.registerFunction(vehicleType, "releaseBales", BaleCollector.releaseBales)
    SpecializationUtil.registerFunction(vehicleType, "setIsBalePositionActive", BaleCollector.setIsBalePositionActive)
end;

function BaleCollector.registerOverwrittenFunctions(vehicleType)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getIsFoldAllowed", BaleCollector.getIsFoldAllowed)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getFillLevelInformation",   BaleCollector.getFillLevelInformation)
end;

function BaleCollector.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", BaleCollector)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", BaleCollector)
	SpecializationUtil.registerEventListener(vehicleType, "onReadStream", BaleCollector)
	SpecializationUtil.registerEventListener(vehicleType, "onWriteStream", BaleCollector)
	SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", BaleCollector)
	SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", BaleCollector)
	SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", BaleCollector)
end;

function BaleCollector:onLoad(savegame)
	self.spec_BaleCollector = {}
	local spec = self.spec_BaleCollector;
	spec.actionEvents = {}

	local key = "vehicle.baleCollector";

	spec.minBaleWidth = MathUtil.round(getXMLFloat(self.xmlFile, key.."#minBaleWidth"), 2);
	spec.maxBaleWidth = MathUtil.round(getXMLFloat(self.xmlFile, key.."#maxBaleWidth"), 2);
	spec.minBaleHeight = MathUtil.round(getXMLFloat(self.xmlFile, key.."#minBaleHeight"), 2);
	spec.maxBaleHeight = MathUtil.round(getXMLFloat(self.xmlFile, key.."#maxBaleHeight"), 2);
	spec.minBaleLength = MathUtil.round(getXMLFloat(self.xmlFile, key.."#minBaleLength"), 2);
	spec.maxBaleLength = MathUtil.round(getXMLFloat(self.xmlFile, key.."#maxBaleLength"), 2);

	spec.baleTriggerNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, key.."#triggerNode"), self.i3dMappings);
	spec.dynamicMountRootNode = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, key.."#rootNode"), self.i3dMappings);
	addTrigger(spec.baleTriggerNode, "baleTriggerCallback", self);
	spec.balesInTrigger = {}
	spec.dynamicMountedObjects = {}
	spec.numBalePositionsInUse = 0;
	spec.numBalesInTrigger = 0;
	spec.forceAcceleration = Utils.getNoNil(getXMLFloat(self.xmlFile, key.."#forceAcceleration"), math.huge);

	spec.balePositions = {}
	local num = 0;
	while true do
		local positionKey = key..".positions.position("..tostring(num)..")";
		if not hasXMLProperty(self.xmlFile, positionKey) then
			break;
		end;

		local balePosition = {}

		balePosition.node = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, positionKey.."#node"), self.i3dMappings);
		balePosition.animationName = getXMLString(self.xmlFile, positionKey.."#animationName");
		balePosition.isActive = false;

		table.insert(spec.balePositions, balePosition);
		num = num + 1;
	end;

	spec.actionEventDirtyFlag = self:getNextDirtyFlag();
	spec.balePositionsDirtyFlag = self:getNextDirtyFlag();
end;

function BaleCollector:onDelete()
	local spec = self.spec_BaleCollector;
	removeTrigger(spec.baleTriggerNode);

	if self.isServer then
		self:releaseBales();
	end;
end;

function BaleCollector:onReadStream(streamId, connection)
	local spec = self.spec_BaleCollector;

	g_inputBinding:setActionEventActive(spec.collectBalesActionEventId, streamReadBool(streamId));
	g_inputBinding:setActionEventActive(spec.releaseBalesActionEventId, streamReadBool(streamId));

	for a=1, #spec.balePositions do
		self:setIsBalePositionActive(a, streamReadBool(streamId));
	end;
end;

function BaleCollector:onWriteStream(streamId, connection)
	local spec = self.spec_BaleCollector;

	streamWriteBool(streamId, spec.numBalesInTrigger > spec.numBalePositionsInUse and self:getIsUnfolded());
	streamWriteBool(streamId, spec.numBalePositionsInUse > 0);

	for a=1, #spec.balePositions do
		streamWriteBool(streamId, spec.balePositions[a].isActive);
	end;
end;

function BaleCollector:onReadUpdateStream(streamId, timestamp, connection)
	local spec = self.spec_BaleCollector;

	if streamReadBool(streamId) then
		g_inputBinding:setActionEventActive(spec.collectBalesActionEventId, streamReadBool(streamId));
		g_inputBinding:setActionEventActive(spec.releaseBalesActionEventId, streamReadBool(streamId));
	end;

	if streamReadBool(streamId) then
		for a=1, #spec.balePositions do
			self:setIsBalePositionActive(a, streamReadBool(streamId));
		end;
	end;
end

function BaleCollector:onWriteUpdateStream(streamId, connection, dirtyMask)
   local spec = self.spec_BaleCollector;

	if streamWriteBool(streamId, bitAND(dirtyMask, spec.actionEventDirtyFlag) ~= 0) then
		streamWriteBool(streamId, spec.numBalesInTrigger > spec.numBalePositionsInUse and self:getIsUnfolded());
		streamWriteBool(streamId, spec.numBalePositionsInUse > 0);
	end;

	if streamWriteBool(streamId, bitAND(dirtyMask, spec.balePositionsDirtyFlag) ~= 0) then
		for a=1, #spec.balePositions do
			streamWriteBool(streamId, spec.balePositions[a].isActive);
		end;
	end;
end;

function BaleCollector:getIsBaleAllowed(bale)
	local spec = self.spec_BaleCollector

	local isRoundBale = bale.baleDiameter ~= nil
	if isRoundBale then
		return false
	end

	if bale.baleWidth == nil
			or bale.baleLength == nil
			or bale.baleHeight == nil then
		return false
	end

	local baleWidth = MathUtil.round(Utils.getNoNil(bale.baleWidth, 0), 2)
	local baleHeight = MathUtil.round(Utils.getNoNil(bale.baleHeight, 0), 2)
	local baleLength = MathUtil.round(Utils.getNoNil(bale.baleLength, 0), 2)

	if baleWidth >= spec.minBaleWidth and baleWidth <= spec.maxBaleWidth
			and baleHeight >= spec.minBaleHeight and baleHeight <= spec.maxBaleHeight
			and baleHeight >= spec.minBaleHeight and baleHeight <= spec.maxBaleHeight then
		return true
	end

	return false
end

function BaleCollector:getNearestBalePosition(bx, by, bz)
	local spec = self.spec_BaleCollector;

	local positionIndex = 0;
	local positionDistance = math.huge;

	for a=1, #spec.balePositions do
		if spec.balePositions[a].bale == nil then
			local px, py, pz = getWorldTranslation(spec.balePositions[a].node);
			local dist = MathUtil.vector3Length(bx-px, by-py, bz-pz);
			if dist < positionDistance then
				positionDistance = dist;
				positionIndex = a;
			end;
		end;
	end;

	return positionIndex, positionIndex ~= 0;
end;

function BaleCollector:baleTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
	local spec = self.spec_BaleCollector;

	if self.isServer then
		local bale = g_currentMission:getNodeObject(otherId)
		if bale ~= nil and bale:isa(Bale) and self:getIsBaleAllowed(bale) then
			if onEnter then
				if spec.balesInTrigger[bale] == nil then
					spec.numBalesInTrigger = spec.numBalesInTrigger + 1;
				end;
				spec.balesInTrigger[bale] = Utils.getNoNil(spec.balesInTrigger[bale], 0)+1;
			end;
			if onLeave then
				spec.balesInTrigger[bale] = spec.balesInTrigger[bale] - 1;
				if spec.balesInTrigger[bale] == 0 then
					spec.balesInTrigger[bale] = nil;
					spec.numBalesInTrigger = spec.numBalesInTrigger - 1;
				end;
			end;
			g_inputBinding:setActionEventActive(spec.collectBalesActionEventId, spec.numBalesInTrigger > spec.numBalePositionsInUse and self:getIsUnfolded());
			self:raiseDirtyFlags(spec.actionEventDirtyFlag);
		end;
	end;
end;

function BaleCollector:addDynamicMountedObject(object)
	local spec = self.spec_BaleCollector;
	spec.dynamicMountedObjects[object] = object;
end;

function BaleCollector:removeDynamicMountedObject(object, isDeleting)
	local spec = self.spec_BaleCollector;
	spec.dynamicMountedObjects[object] = nil;
end;

function BaleCollector.actionEventCollectBales(self, actionName, inputValue, callbackState, isAnalog)
	if self.isServer then
		self:collectBales();
	else
		g_client:getServerConnection():sendEvent(BaleCollectorEvent:new(self, true));
	end;
end;

function BaleCollector.actionEventReleaseBales(self, actionName, inputValue, callbackState, isAnalog)
	if self.isServer then
		self:releaseBales();
	else
		g_client:getServerConnection():sendEvent(BaleCollectorEvent:new(self, false));
	end;
end;

function BaleCollector:collectBales()
	local spec = self.spec_BaleCollector;

	if self.isServer and self:getIsUnfolded() then
		for bale, v in pairs(spec.balesInTrigger) do
			if bale ~= nil and bale.nodeId ~= nil and entityExists(bale.nodeId) then
				local bx, by, bz = getWorldTranslation(bale.nodeId);
				local positionIndex, foundPlace = self:getNearestBalePosition(bx, by, bz);
				if foundPlace and spec.dynamicMountedObjects[bale] == nil then
					local balePosition = spec.balePositions[positionIndex];
					spec.balesInTrigger[bale] = nil; -- bale will be detected again after unmount, this is to avoid double entries
					spec.numBalesInTrigger = spec.numBalesInTrigger - 1;

					local rx, ry, rz = localDirectionToWorld(balePosition.node, 0, 0, 1);
					local bx, by, bz = localDirectionToWorld(bale.nodeId, 0, 0, 1);

					local scalar = rx*bx+rz*bz;
					local lengthA = MathUtil.vector2Length(rx, rz);
					local lengthB = MathUtil.vector2Length(bx, bz);

					local angle = math.acos(scalar/(lengthA*lengthB));

					local baleMountRy = 0;
					if angle > (0.5*math.pi) and (lengthA*lengthB) > 0 then
						baleMountRy = math.pi;
					end;

					bale:mount(self, balePosition.node, 0,0,0, 0,baleMountRy,0);
					bale:unmount();

					if bale.dynamicMountJointIndex == nil then
						bale:mountDynamic(self, spec.dynamicMountRootNode, balePosition.node, DynamicMountUtil.TYPE_AUTO_ATTACH_XYZ, spec.forceAcceleration);
						self:addDynamicMountedObject(bale);
						balePosition.bale = bale;
						self:setIsBalePositionActive(positionIndex, true);
					end;
				end;
			end;
		end;
	end;
end;

function BaleCollector:releaseBales()
	local spec = self.spec_BaleCollector;

	if self.isServer then
		for a=1, #spec.balePositions do
			local balePosition = spec.balePositions[a];
			if balePosition.bale ~= nil then
				balePosition.bale:unmountDynamic();
				self:removeDynamicMountedObject(balePosition.bale, false);
				balePosition.bale = nil;
				self:setIsBalePositionActive(a, false);
			end;
		end;
	end;
end;

function BaleCollector:setIsBalePositionActive(positionIndex, isActive)
	local spec = self.spec_BaleCollector;

	local balePosition = spec.balePositions[positionIndex];

	if balePosition.isActive ~= isActive then
		balePosition.isActive = isActive;
		if isActive then
			self:playAnimation(balePosition.animationName, 1, 0, true);
			spec.numBalePositionsInUse = spec.numBalePositionsInUse + 1;
		else
			self:playAnimation(balePosition.animationName, -1, 1, true);
			spec.numBalePositionsInUse = math.max(spec.numBalePositionsInUse - 1, 0);
		end;
	end;

	if self.isServer then
		g_inputBinding:setActionEventActive(spec.releaseBalesActionEventId, spec.numBalePositionsInUse > 0);
		g_inputBinding:setActionEventActive(spec.collectBalesActionEventId, spec.numBalesInTrigger > spec.numBalePositionsInUse and self:getIsUnfolded());
		self:raiseDirtyFlags(spec.balePositionsDirtyFlag);
		self:raiseDirtyFlags(spec.actionEventDirtyFlag);
	end;
end;

function BaleCollector:getIsFoldAllowed(superFunc)
	local spec = self.spec_BaleCollector;

	if spec.numBalePositionsInUse > 0 then
		return false;
	end;
	return superFunc(self);
end;

function BaleCollector:getFillLevelInformation(superFunc, fillLevelInformation)
	superFunc(self, fillLevelInformation);

	local spec = self.spec_BaleCollector;

	for bale, _ in pairs(spec.dynamicMountedObjects) do
		if bale.getFillLevel ~= nil and bale.getFillType ~= nil then
			local fillTypeExists = false;
			for _, info in pairs(fillLevelInformation) do
				if info.fillType == bale:getFillType() then
					info.fillLevel = info.fillLevel + bale:getFillLevel();
					info.capacity = info.capacity + bale:getFillLevel();
					fillTypeExists = true;
				end;
			end;
			if not fillTypeExists then
				local info = {}
				info.fillType = bale:getFillType();
				info.fillLevel = bale:getFillLevel();
				info.capacity = info.fillLevel;
				table.insert(fillLevelInformation, info);
			end;
		end;
	end;
end;

function BaleCollector:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
	if self.isClient then
        local spec = self.spec_BaleCollector;
		self:clearActionEventsTable(spec.actionEvents);
		if isActiveForInputIgnoreSelection then
			local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.ACTIVATE_OBJECT, self, BaleCollector.actionEventCollectBales, false, true, false, true, nil)
			g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL);
			g_inputBinding:setActionEventText(actionEventId, g_i18n:getText("function_collectBales", self.customEnvironment));
			spec.collectBalesActionEventId = actionEventId;
			g_inputBinding:setActionEventActive(spec.collectBalesActionEventId, spec.numBalesInTrigger > spec.numBalePositionsInUse and self:getIsUnfolded());

			local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.IMPLEMENT_EXTRA3, self, BaleCollector.actionEventReleaseBales, false, true, false, true, nil)
			g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL);
			g_inputBinding:setActionEventText(actionEventId, g_i18n:getText("function_releaseBales", self.customEnvironment));
			spec.releaseBalesActionEventId = actionEventId;
			g_inputBinding:setActionEventActive(spec.releaseBalesActionEventId, spec.numBalePositionsInUse > 0);
		end;
	end;
end;